package CoreSim::Chord;

use strict;

use CoreSim::Chord::ChordRing;  
use CoreSim::Chord::ChordMetrics;
use CoreSim::Utilities::IPUtilities;


sub new(){
	my $invocant = shift;
	my ( $topology, $maxNodesInChordRing, $maxNodesInFingerTable, $readFingerTablesFromFile, $startPosition, $endPosition,
	$sortedAndFilteredPrefixListFile, $resolverType, $outputFilePath, $_DEBUG) = @_;
	
	my $class = ref($invocant) || $invocant;
	my $ipUtility = CoreSim::Utilities::IPUtilities->new();
	
	my $self = {
		"topology" => $topology,
		"maxNodesInChordRing" => $maxNodesInChordRing,
		"maxNodesInFingerTable" => $maxNodesInFingerTable,
		"readFingerTablesFromFile" => $readFingerTablesFromFile,
		"startPosition" => $startPosition,
		"endPosition" => $endPosition,
		"sortedAndFilteredPrefixListFile" => $sortedAndFilteredPrefixListFile,
		"ipUtility" => $ipUtility,
		"resolverType" => $resolverType,
		"outputFilePath" => $outputFilePath,
		"_DEBUG" => $_DEBUG,
	};
	
	&initialize($self);
	
	bless($self, $class);
	
	return $self;
}

sub initialize(){
	my $self = shift;
	
	my $topology = $self->{"topology"};
	my $maxNodesInChordRing = $self->{"maxNodesInChordRing"};
	my $maxNodesInFingerTable = $self->{"maxNodesInFingerTable"};
	my $readFingerTablesFromFile = $self->{"readFingerTablesFromFile"};
	my $ipUtility = $self->{"ipUtility"};
	my $resolverType = $self->{"resolverType"};
	my $debug = $self->{"_DEBUG"};
	
	my $startPosition = $self->{"startPosition"};
	my $endPosition = $self->{"endPosition"};
	my $sortedAndFilteredPrefixListFile = $self->{"sortedAndFilteredPrefixListFile"};
	my $outputFilePath = $self->{"outputFilePath"};
	
	my $chordRing = CoreSim::Chord::ChordRing->new($topology,$maxNodesInChordRing, $maxNodesInFingerTable,$readFingerTablesFromFile, 
													$startPosition, $endPosition,$sortedAndFilteredPrefixListFile, $ipUtility,$debug );
	my $chordMetrics = CoreSim::Chord::ChordMetrics->new($chordRing, $topology, $resolverType, $ipUtility,$outputFilePath, $debug);
														
	$self->{"chordRing"} = $chordRing;
	$self->{"chordMetrics"} = $chordMetrics; 
	
	
}

#returns the time needed for a DHT lookup( for a given ID from a given node )
sub getDHTLookupDelayFromPOP(){
	my $self = shift;
	my ($startNode, $chordID) = @_;
	
	return $self->{"chordMetrics"}->getDHTLookupDelayFromPOP($startNode, $chordID);
}

sub getDHTLookupDelayFromChordID(){
	my $self = shift;
	my ($startChordID, $ip) = @_;
	
	return $self->{"chordMetrics"}->getDHTLookupDelayFromChordID($startChordID, $ip);
}

sub dropFingerTablesToFile(){
	my $self = shift;
	my $filePath = shift;
	$self->{"chordRing"}->dropFingerTablesToFile($filePath);
}

sub dropFancyFingerTablesToFile(){
	my $self = shift;
	my $filePath = shift;
	$self->{"chordRing"}->dropFancyFingerTablesToFile($filePath);
}

sub dropSortedAndFilteredPrefixesToFile(){
	my $self=shift;
	my $outputFile = shift;
	
	$self->{"chordRing"}->dropSortedAndFilteredPrefixesToFile($outputFile);
}

sub getChordIDForPrefix(){
	my $self = shift;
	
	my $prefix = shift;
	my $ipUtility = $self->{"ipUtility"};
	return $ipUtility->getChordIDForPrefix($prefix);
}

sub getResolverType(){
	my $self=shift;
	
	my $chordMetrics = $self->{"chordMetrics"};
	return 	$chordMetrics->getResolverType();
}

sub setNodeIP(){
	my $self=shift;
	my ($chordNodeID, $ip) = @_;
	
	my $chordNode = $self->{"chordRing"}->getNodeWithID($chordNodeID);
	$chordNode->setIP($ip);	
}

sub showFingerTablesPrefixesForChordID(){
	my $self=shift;
	my ($chordID) = @_;
	
	$self->{"chordRing"}->showFingerTablesPrefixesForChordID($chordID);
}

#common name for all the mapping systems for the function which computes the latency between an associated node (here the chordID)
#and a searched IP
sub getLookupDelay(){
	my $self= shift;
	my ($startChordID, $ip) = @_;
	
	return $self->{"chordMetrics"}->getDHTLookupDelayFromChordID($startChordID, $ip);
}

#each mapping system has this function and it uses it to transform a prefix (given as parameter) into an appropriate form
#to be used by the getLookupDelayFunction. Here the prefix is transformed into a ChordID
sub getAssociatedNodeForPrefix(){
	my $self=shift;
	my $prefix = shift;
	
	return &getChordIDForPrefix($self,$prefix);
}

#return 1 
1