#################################################################################
# Copyright (c) 2009 Florin Coras.                                              #
# Copyright (c) 2009 Universitat Politecnica de Catalunya.                      #
#                                                                               #
# Permission is hereby granted, free of charge, to any person obtaining a       #
# copy of this software and associated documentation files (the "Software"),    #	
# to deal in the Software without restriction, subject to the conditions        #
# listed in the LICENSE file. These conditions include: you must                #
# preserve this copyright notice, and you cannot mention the copyright          #
# holders in advertising related to the Software without their permission.      #
# The Software is provided WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED. This       #
# notice is a summary of the LICENSE file; the license in that file is          #
# legally binding.                                                              #
#                                                                               #
#################################################################################

package CoreSim::Chord::ChordMetrics;


use strict;

#use CoreSim::Chord::ChordNode;
#use CoreSim::Chord::FingerTable;

sub new(){
	
	my $invocant = shift;
	my ($chordRing, $topology, $resolverType, $ipUtility,$outputFilePath, $_DEBUG) = @_;
	
	my $class = ref($invocant) || $invocant;
	
	my $numberOfQueriesHash={};
	
	my $self = {
		"chordRing" => $chordRing,
		"topology" => $topology,
		"resolverType" => $resolverType,
		"ipUtility" => $ipUtility,
		"numberOfQueriesHash" => $numberOfQueriesHash,
		"outputFilePath" => $outputFilePath,
		"_DEBUG"=> $_DEBUG,
	};
	
	
	bless($self, $class);
#	$self->{"_DEBUG"} = 3;
	return $self;
}

#returns lookup delay of a given chordID starting in a given POP
sub getDHTLookupDelayFromPOP(){

}

sub getDHTLookupDelayFromChordID(){
	my $self = shift;
	my ($startChordID, $searchedIP) = @_;
	
	my $startNode = $self->{"chordRing"}->getNodeWithID($startChordID);
	
	if (!defined($startNode)){
		printf("\n(EE)(gDLDCI) The node with chordID %s does not exist in ChordRing!\n\n", $startChordID);
		return undef;
	}

	
	return &getDHTLookupDelay($self, $startNode, $searchedIP );
}

sub getDHTLookupDelay(){
	
	my $self = shift;
	my ($startNode, $searchedIP) = @_;
	
	my $ipUtility = $self->{"ipUtility"};
	
	my $searchedChordID = $ipUtility->getIPToNumber($searchedIP);
	
	printf("\n(gDLDCI) The searched IP %s has the ChordID=%s\n", $searchedIP, $searchedChordID) if ($self->{"_DEBUG"});
	
	
	#find the successor and the ChordPath for the given ChordID
	my ( $successor, $path) = &findSuccessorForID($self, $startNode, $searchedChordID); 
	
	my $successorPrefix = $successor->getPrefix();
	my $pathLength = scalar(@$path);
	
	#check that the successor for the ChordID is actually the responsible router for the searchedIP
	if ($ipUtility->ipInsidePrefix($searchedIP, $successorPrefix)){
		
		my $latency;
		my $nodeLatencyList;
		my $numberOfQueries;
		my $numberOfAnsweredQueries;
		
		#the path is composed by more than one node
		if ($pathLength > 1){
			($latency, $nodeLatencyList, $numberOfQueries, $numberOfAnsweredQueries) = &computePathLatency($self, $path);
			
			printf("\n(gDLD) In order to find the latency implied by the Chord search for the ChordID %s %s queries have been implied\n",
				$searchedChordID, $numberOfQueries) if ($self->{"_DEBUG"});
				
		#the start node is responsible for the searched ChordID
		}else{
			$latency = 0;
			$nodeLatencyList = [];
			$numberOfQueries = 0;
			$numberOfAnsweredQueries = 0;
			
		}

		
		
		#update the numberOfQueriesHash for all the nodes in the path. This is done only for valid queries
		my $numberOfQueriesHash = $self->{"numberOfQueriesHash"};
		&updateNumberOfQueriesHash($numberOfQueriesHash, $path);
		
		
		#organize the return list
		my $status = "ok";
		my $destPOP = $successor->getIP();
		my $destAS = $successor->getAS();
		my $destPrefix = $successor->getPrefix();
		my $lowerBoundLatency = $latency;
		my $nbOfPeers = scalar(@$path) -1;
		
		return ($status, $destPOP, $destAS, $destPrefix, $lowerBoundLatency, $nbOfPeers, $nodeLatencyList);
	
	#if the successor is not responsible for the searchedIP the ITR is "announced" to discard the packet
	}else{
		my $status = "discard";
		my $destPOP = $successor->getIP();
		my $destAS = $successor->getAS();
		my $destPrefix = $successor->getPrefix();
		return ($status, $destPOP, $destAS, $destPrefix);
	}

}

#find the successor for a given id if we start in a given node. It returns the node and the path followed
sub findSuccessorForID(){

	my ($self, $startNode, $searchedChordID) = @_;
	
	my $chordPredecessorNode;
	my $chordPath;
	
	
	#if this node is responsible for the searchedChordID return 0 -> no latency implied. This is a special case!
	if ( &idIsBetweenNodes($self,$searchedChordID, $startNode->getPredecessor->getChordID(), $startNode->getChordID()) ){
		
		printf("\n(gDLD) The startNode with chordID %s has the key %s. The search latency is 0.\n", 
				$startNode->getChordID(), $searchedChordID) if ($self->{"_DEBUG"});
		
		my $successor = $startNode;
		push (@$chordPath, $startNode);
		 
		return ($successor, $chordPath);
	}
	
	
	printf("\n(FS) The start POP is %s with the IP %s and the ChordID %s\n",
		 $startNode->getPOP(), $startNode->getIP(), $startNode->getChordID()) if ($self->{"_DEBUG"}>=2);
	
	#find the predecessor of the searchedChordNode
	($chordPredecessorNode, $chordPath) = &findPredecessorForID($self, $startNode, $searchedChordID);
	
	#add the successor to the chord path
	my $successor = $chordPredecessorNode->getSuccessor();
	push(@$chordPath, $successor);
	
	printf("(FS) We found successor for ChordID %s. It is POP %s with the IP %s and the ChordID %s\n",
			$searchedChordID, $successor->getPOP(), $successor->getIP(), $successor->getChordID()) if ($self->{"_DEBUG"}>=2);
	
	
	if ($self->{"_DEBUG"}){
		printf("(FS) The chordID path followed when searching for ChordID %s: ", $searchedChordID);
		foreach my $chordNode (@$chordPath){
			printf("%s ",$chordNode->getChordID());	
		}
		printf("\n");
	}
	
	return ($successor, $chordPath) ;
}

sub findPredecessorForID(){
	
	my ($self, $startNode, $searchedChordID) = @_;
	
	my $currentNode = $startNode;
	my @chordPath;
	
	
	push(@chordPath, $currentNode);
	
	#the case when we search for a predecessor and the node we start with is exactly the predecessor
	if ( &idIsBetweenNodes($self, $searchedChordID, $currentNode->getChordID(), $currentNode->getSuccessor()->getChordID()) ){
		
		push (@chordPath, $currentNode->getSuccessor());
		
		printf("(FP) The start node %s is the predecessor for the searched ChordID %s\n", $currentNode->getChordID(), $searchedChordID)
		if ($self->{"_DEBUG"}>=2);
		
		return ($currentNode,\@chordPath);	
	}
	
	#while we do not find a node for which the chordID lies between the node itself and its successor, change the currentNode such that 
	#it is the closest preceding finger of the searched chordID
	while ( ( ! &idIsBetweenNodes($self, $searchedChordID, $currentNode->getChordID(), $currentNode->getSuccessor()->getChordID()) ) && 
		( $currentNode->getChordID() != $searchedChordID )){
		
		printf("\n(FP) Ask the POP %s with ChordID %s\n", $currentNode->getPOP(), $currentNode->getChordID() ) if ($self->{"_DEBUG"}>=2);
		
		$currentNode = &closestPrecedingFinger($self, $currentNode, $searchedChordID);
		push(@chordPath, $currentNode);
		
	}	
	
	
	return ($currentNode,\@chordPath);
	
}

#returns the closest preceding finger
sub closestPrecedingFinger(){
	my $self = shift;
	my ($node, $id) = @_;
	
	my $fingerTable = $node->getFingerTable();
	my $fingerTableSize = $fingerTable->getFingerTableSize();
	
	
	for (my $i=$fingerTableSize-1; $i>=0; $i-=1){
		my $fingerNode = $fingerTable->getNodeForFinger($i);
		
		printf("(CPF) Checking the finger number %s, which is the POP %s with the IP %s and ChordID =%s\n", 
				$i, $fingerNode->getPOP(),$fingerNode->getIP(), $fingerNode->getChordID) if ($self->{"_DEBUG"}>=2);
		
		#if the fingerNode (chordID) is situated between the node and the id than we return this node
		if ( &fingerIsBetweenNodeAndID($self, $fingerNode->getChordID(), $node->getChordID(), $id) ){
			printf("(CPF) The closest preceding finger has the number %s and is the POP %s with the IP %s and the ChordID %s\n", 
					$i, $fingerNode->getPOP(), $fingerNode->getIP(), $fingerNode->getChordID()) if ($self->{"_DEBUG"}>=2);
			return $fingerNode;
		}
		
	}
}

sub computePathLatency(){
	my ($self, $path) = @_;
	
	my $topology = $self->{"topology"};
	
	#the resolverType can be recursive(0) or iterative(1)
	my $resolverType = $self->{"resolverType"};
	
	my $node1 = $path->[0];
	my $latency = 0;
	my @nodeLatencyList;
	
	my $numberOfQueries = 0;
	my $numberOfAnsweredQueries = 0;
	
	#add the first node to the nodeLatencyList
	push(@nodeLatencyList, $node1->getIP());
	
	#recursive type of resolver
	if ($resolverType == 0){
		
		for (my $i=1; $i< scalar( @$path); $i++){
			my $node2 = $path->[$i];
			my $latencyQuery = $topology->getLatencyBetweenPOPIPs1($node1->getIP(),$node2->getIP() );
			
			$numberOfQueries++;
		
			$numberOfAnsweredQueries++;
			$latency += $latencyQuery;
				
			#create nodeLatencyList
			push(@nodeLatencyList, $latencyQuery );
			push(@nodeLatencyList,$node2->getIP());
			
			$node1 = $node2;
		}
		
		#after changes in the topologyMetrics, the result of a latencyQuery will always be valid
		my $returnLatency = $topology->getLatencyBetweenPOPIPs1($path->[0]->getIP(), $path->[-1]->getIP());

		$numberOfAnsweredQueries++;
		$latency += $returnLatency;		

		
		#add the return latency to the nodeLatencyList
		push(@nodeLatencyList, $returnLatency);
		push(@nodeLatencyList, $path->[0]->getIP());
		$numberOfQueries++;
		
		return ($latency, \@nodeLatencyList, $numberOfQueries, $numberOfAnsweredQueries);
	
	#iterative type of resolver	
	}else{
		
		for (my $i=1; $i< scalar( @$path); $i++){
			my $node2 = $path->[$i];
			my $latencyQuery = $topology->getLatencyBetweenPOPIPs1($node1->getIP(),$node2->getIP() );
			
			$numberOfQueries++;
		
			$numberOfAnsweredQueries++;
			$latency += 2*$latencyQuery;
		
			#create nodeLatencyList
			push(@nodeLatencyList,$node2->getIP());
			push(@nodeLatencyList, $latencyQuery );
		}
		
		return ($latency,\@nodeLatencyList, $numberOfQueries, $numberOfAnsweredQueries);
		
	}
		
}

sub updateNumberOfQueriesHash(){
	my ($numberOfQueriesHash, $path) = @_;
	
	my $pathLength = scalar(@$path);
	
	for (my $i=0;$i<$pathLength; $i++){
		$numberOfQueriesHash->{$path->[$i]->getChordID()} += 1;
	}
}

#returns 1 if the node1_ID<id<= node2_ID 
sub idIsBetweenNodes(){
	my $self = shift;
	my ($id, $chordID1, $chordID2) = @_;
	
	if ($chordID1<$chordID2){
		if ( ($chordID1<$id) && ($id<=$chordID2) ){
			return 1;
		}else{
			return 0;
		}
		
	#this is the case when a jump from the last node to the first is needed
	}else{
		
		#get nodesInChordRing, also the maximum number of nodes in the ring 
		#ATTENTION! the 0 in the comparison might prove to be a bad choice!
		my $nodesInChordRing = $self->{"chordRing"}->getNodesInChordRing();
		
		if ( (($chordID1<$id) && ($id<= $nodesInChordRing)) || (( 0<=$id) && ($id< $chordID2) ) ){
			return 1;
		}else {
			return 0;
		}
	}
}

#this is different from idIsBetweenNodes, check the inequality
sub fingerIsBetweenNodeAndID(){
	
	my $self = shift;
	
	my($finger, $node, $id) = @_;
	
	if ($node < $id){
		if( ($node< $finger) && ($finger<$id) ){
			return 1;
		}else{
			return 0;
		}
	}else{
		my $nodesInChordRing = $self->{"chordRing"}->getNodesInChordRing();

		if ( (($node < $finger) &&($finger <= $nodesInChordRing)) || ( (0<=$finger) && ($finger < $id) ) ){
			return 1;		
		}else{
			return 0;
		}
	}
}

sub getResolverType(){
	my $self=shift;
	
	return $self->{"resolverType"};
}

sub DESTROY(){
	my $self=shift;
	
	#write chordNodesStats
	my $outputFilePath = $self->{"outputFilePath"};
	my $outputFile = "$outputFilePath/chordNodesStats.txt";
	my $numberOfQueriesHash = $self->{"numberOfQueriesHash"};
	my $chordRing = $self->{"chordRing"};
	
	open(OFILE, "> $outputFile") or die printf("Error while opening file %s to print chord node stats\n", $outputFile);
	
	my @chordIDs = keys %$numberOfQueriesHash;
	
	foreach my $chordID (@chordIDs){
		my $chordNode = $chordRing->getNodeWithID($chordID);
		my $prefix = $chordNode->getPrefix();
		my $asNumber = $chordNode->getAS();
		my $queriesCount = $numberOfQueriesHash->{$chordID};
		
		printf(OFILE "%s %s %s %s\n", $chordID, $prefix, $asNumber, $queriesCount );
	}
}

#return 1 at the end
1